import json
import hashlib

def dict_to_hash(d: dict) -> str:
    hasher = hashlib.sha256()

    hasher.update(json.dumps(d).encode("utf-8"))

    return hasher.hexdigest()

class CacheKeyMissingError(Exception):
    pass

class DiskCache:
    """Store and retrieve keys from JSON object on disk."""

    def __init__(self, path):
        self.path = path

    def _read_cache(self) -> dict:
        try:
            with open(self.path, "r") as f:
                return json.load(f)
        except (FileNotFoundError, json.JSONDecodeError):
            return {}


    def _write_cache(self, cache: dict) -> None:
        with open(self.path, "w") as f:
            json.dump(cache, f, indent=4, sort_keys=True)


    def get(self, key: str, inputs: dict) -> object:
        cache = self._read_cache()

        try:
            entry = cache[key]

            if entry.get("inputs_hash") != dict_to_hash(inputs):
                raise CacheKeyMissingError()

            return entry["value"]
        except KeyError:
            raise CacheKeyMissingError()

    def set(self, key: str, inputs: dict, value: object) -> None:
        cache = self._read_cache()

        cache[key] = dict(
            inputs_hash=dict_to_hash(inputs),
            value=value,
        )

        self._write_cache(cache)



def main(path: str):
    #<<filenames = a list of filenames under path>>
    import os
    filenames = [os.path.join(path, f) for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
    #<</>>

    for fn in filenames:
        #<<size = size of fn in bytes>>
        size = os.path.getsize(fn)
        #<</>>
        print(fn, size)

#<<use argparse and call main>>
import argparse
parser = argparse.ArgumentParser()
parser.add_argument("path", type=str)
args = parser.parse_args()
main(args.path)
#<</>>



import re
import logging

from dataclasses import dataclass
from itertools import chain

from parsimonious.nodes import (
    Node,
    NodeVisitor,
)
from parsimonious.grammar import Grammar

from maccarone.openai import CachedChatAPI

logger = logging.getLogger(__name__)

class Piece:
    pass

@dataclass
class PresentPiece(Piece):
    text: str

@dataclass
class MissingPiece(Piece):
    indent: str
    guidance: str
    inlined: str | None = None

grammar = Grammar(
    r"""
    maccarone = human_source maccarone_chunk*
    maccarone_chunk = snippet human_source?

    snippet = snippet_open (ai_source snippet_close)?
    snippet_open = snippet_open_single / snippet_open_multi
    snippet_open_single = guidance_open guidance_inner ">>" nl
    snippet_open_multi = guidance_open nl guidance_lines guidance_close
    snippet_close = ws "#<</>>" nl

    guidance_open = ws "#<<"
    guidance_close = ws "#>>" nl
    guidance_line = ws "#" guidance_inner nl
    guidance_lines = guidance_line+
    guidance_inner = ~"((?!>>).)*"

    human_source = source_line*
    ai_source = source_line*
    source_line = !(guidance_open / guidance_close / snippet_close) ws ~".*" nl?

    ws = ~"[ \t]*"
    nl = ws ~"[\r\n]"
    """
)

@dataclass
class GuidanceOpen:
    indent: str

@dataclass
class Guidance:
    text: str

@dataclass
class SnippetOpen:
    indent: str
    guidance: str

class RawSourceVisitor(NodeVisitor):
    def generic_visit(self, node: Node, visited_children: list[Node]):
        return visited_children or node

    def visit_maccarone(self, node: Node, visited_children: list):
        (first_source, chunks) = visited_children

        return [first_source] + list(chain(*chunks))

    def visit_maccarone_chunk(self, node: Node, visited_children: list):
        (snippet, source) = visited_children

        if isinstance(source, list):
            source_list = source
        else:
            source_list = []

        return [snippet] + source_list

    def visit_snippet(self, node: Node, visited_children: list):
        (snippet_open, quantified_source) = visited_children

        if isinstance(quantified_source, list):
            ((source, _),) = quantified_source
        else:
            source = None

        return MissingPiece(
            indent=snippet_open.indent,
            guidance=snippet_open.guidance,
            inlined=source,
        )

    def visit_snippet_open(self, node: Node, visited_children: list):
        (single_or_multi,) = visited_children

        return single_or_multi

    def visit_snippet_open_single(self, node: Node, visited_children: list):
        (guidance_open, guidance, _, _) = visited_children

        return SnippetOpen(
            indent=guidance_open.indent,
            guidance=guidance.text,
        )

    def visit_snippet_open_multi(self, node: Node, visited_children: list):
        (guidance_open, _, guidance, _) = visited_children

        return SnippetOpen(
            indent=guidance_open.indent,
            guidance=guidance.text,
        )

    def visit_guidance_open(self, node: Node, visited_children: list):
        (ws, _) = visited_children

        return GuidanceOpen(indent=ws.text)

    def visit_guidance_line(self, node: Node, visited_children: list):
        (_, _, guidance_inner, _) = visited_children

        return guidance_inner

    def visit_guidance_lines(self, node: Node, visited_children: list):
        return Guidance(
            text="\n".join(g.text for g in visited_children)
        )

    def visit_guidance_inner(self, node: Node, visited_children: list):
        return Guidance(text=node.text)

    def visit_human_source(self, node: Node, visited_children: list):
        return PresentPiece(text=node.text)

    def visit_ai_source(self, node: Node, visited_children: list):
        return node.text

def raw_source_to_pieces(input: str) -> list[Piece]:
    tree = grammar.parse(input)
    visitor = RawSourceVisitor()
    pieces = visitor.visit(tree)

    return pieces

def raw_pieces_to_tagged_input(raw_pieces: list[Piece]) -> str:
    tag_source = ""
    id = 0

    for piece in raw_pieces:
        match piece:
            case PresentPiece(text):
                tag_source += text
            case MissingPiece(indent, guidance):
                tag_source += f'# <write_this id="{id}">\n{indent}# {guidance}\n{indent}# </>'
                id += 1
            case _:
                raise TypeError("unknown piece type", piece)

    logger.debug("tagged input ↓\n%s", tag_source)

    return tag_source

def tagged_input_to_tagged_output(tagged_input: str, chat_api: CachedChatAPI) -> str:
    system_prompt = """
You are an expert programmer working on contract. Your client has written a partial program, but left pieces for you to complete. They have marked those with `<write_this>` tags inside Python comments, e.g.:

```
def add_two_numbers(x, y):
    # <write_this id="0">
    # add the two numbers
    # </>

# <write_this id="1">
# add two numbers from command line args, using argparse
# </>
```

You should produce a document that provides a `<completed>` tag for each missing piece, e.g.:

```
<completed id="0">
return x + y
</>
<completed id="1">
import argparse
parser = argparse.ArgumentParser()
parser.add_argument("x", type=int)
parser.add_argument("y", type=int)
args = parser.parse_args()
return add_two_numbers(args.x, args.y)
</>
```

This formatting is very important. The client uses a custom tool to process your work product, and their tool requires this format. Follow this format exactly and do not copy anything outside a `<write_this>` tag.
"""
    chat_messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": tagged_input},
    ]
    tagged_output = chat_api.complete_chat("tagged_input_to_tagged_output", chat_messages)

    logger.debug("tagged output ↓\n%s", tagged_output)

    return tagged_output

def tagged_output_to_completed_pieces(tagged_output: str) -> dict[int, str]:
    pattern = re.compile(r'<completed id="(?P<id>\d+)">\n(?P<content>.+?)</>', re.DOTALL)
    matches = pattern.finditer(tagged_output)
    completed = {int(m.group("id")): m.group("content") for m in matches}

    return completed

def pieces_to_final_source(
        raw_pieces: list[Piece],
        completed_pieces: dict[int, str],
    ) -> str:
    id = 0
    final_source = ""

    for raw in raw_pieces:
        match raw:
            case PresentPiece(text):
                final_source += text

            case MissingPiece(indent, guidance):
                final_source += f"{indent}#<<{guidance}>>\n"
                completed = completed_pieces[id]
                final_source += indent + indent.join(completed.splitlines(True))
                final_source += f"{indent}#<</>>\n"
                id += 1

            case _:
                raise TypeError("unknown piece type", raw)

    logger.debug("final source ↓\n%s", final_source)

    return final_source

def preprocess_maccarone(
        raw_source: str,
        chat_api: CachedChatAPI,
    ) -> str:
    raw_pieces = raw_source_to_pieces(raw_source)
    tagged_input = raw_pieces_to_tagged_input(raw_pieces)
    tagged_output = tagged_input_to_tagged_output(tagged_input, chat_api)
    completed_pieces = tagged_output_to_completed_pieces(tagged_output)
    final_source = pieces_to_final_source(
        raw_pieces,
        completed_pieces,
    )

    return final_source



import os
import logging

from typing import (
    Callable,
    Iterable,
    cast,
)

import openai

from openai import ChatCompletion

from maccarone.caching import (
    DiskCache,
    CacheKeyMissingError,
)

logger = logging.getLogger(__name__)

openai.api_key = os.getenv("OPENAI_API_KEY")
openai.api_base = os.getenv("OPENAI_API_BASE", openai.api_base)

def complete_chat(
        messages: list[dict[str, str]],
        model="gpt-4",
        on_token: Callable[[int], None] = lambda p: None,
    ) -> str:
    helicone_key = os.getenv("HELICONE_API_KEY")

    if helicone_key is None:
        headers = {}
    else:
        headers={"Helicone-Auth": helicone_key}

    responses = cast(
        Iterable[ChatCompletion],
        ChatCompletion.create(
            model=model,
            messages=messages,
            stream=True,
            temperature=0.0,
            headers=headers,
        ),
    )
    completion = ""

    logger.info("completing: %r", messages)

    for (i, partial) in enumerate(responses):
        delta = partial.choices[0].delta

        try:
            completion += str(delta.content)
        except AttributeError as error:
            pass

        on_token(i)

    logger.info("completion: %r", completion)

    return completion

class CachedChatAPI:
    def __init__(self, cache_path: str):
        self.cache = DiskCache(cache_path)

    def complete_chat(
            self,
            chat_name: str,
            messages: list[dict[str, str]],
            model="gpt-4",
        ) -> str:
        try:
            return self.cache.get(chat_name, messages)
        except CacheKeyMissingError:
            completion = complete_chat(messages, model=model)

            self.cache.set(chat_name, messages, completion)

            return completion



import os
import glob
import logging

from argparse import Namespace

from maccarone.openai import CachedChatAPI
from maccarone.preprocessor import preprocess_maccarone

logger = logging.getLogger(__name__)

def preprocess(
        mn_path: str,
        print_: bool,
        write: bool,
        rewrite: bool,
        suffix: str,
    ) -> None:
    # produce Python source
    logger.info("preprocessing %s", mn_path)

    cache_path = mn_path.replace(suffix, ".mn.json")
    chat_api = CachedChatAPI(cache_path)

    mn_source = open(mn_path, "r").read()


    py_source = preprocess_maccarone(mn_source, chat_api)

    if write:
        import re
        py_path = re.sub(f"{suffix}$", ".py", mn_path)


        logger.info("writing %s", py_path)

        if py_path == mn_path:
            raise ValueError("won't overwrite input file", mn_path)
    elif rewrite:
        py_path = mn_path
    else:
        py_path = None

    if py_path is not None:
        with open(py_path, "w") as f:
            f.write(py_source)


    if print_:
        print(py_source, end="")

def main(path: str, print_: bool, write: bool, rewrite: bool, suffix: str) -> None:
    """Preprocess files with Maccarone snippets."""

    if os.path.isdir(path):
        mn_files = glob.glob(
            os.path.join(path, f"**/*{suffix}"),
            recursive=True,
        )
    else:
        mn_files = [path]

    for mn_file in mn_files:
        preprocess(mn_file, print_, write, rewrite, suffix)


def parse_args() -> Namespace:
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("path", help="Path to the file or directory containing Maccarone snippets")
    parser.add_argument("--print", dest="print_", action="store_true", help="Print the preprocessed Python source")
    parser.add_argument("--write", action="store_true", help="Write the preprocessed Python source to a new file")
    parser.add_argument("--rewrite", action="store_true", help="Overwrite the input file with the preprocessed Python source")
    parser.add_argument("--suffix", default=".mn.py", help="Suffix for Maccarone snippet files (default: .mn.py)")
    args = parser.parse_args()
    return args


def script_main():
    logging.basicConfig(level=logging.INFO)

    return main(**vars(parse_args()))

if __name__ == "__main__":
    script_main()




import os.path

from tempfile import TemporaryDirectory

from maccarone.caching import (
    DiskCache,
    CacheKeyMissingError,
)

def test_disk_cache_set_get():
    with TemporaryDirectory() as cache_dir:
        cache = DiskCache(os.path.join(cache_dir, "cache.json"))

        name = "foo"
        inputs = {"a": 1, "b": 2}
        value = {"c": 3, "d": 4}

        try:
            cache.get(name, inputs)
        except CacheKeyMissingError:
            pass
        else:
            assert False, "expected CacheKeyMissingError"

        cache.set(name, inputs, value)

        assert cache.get(name, inputs) == value



import os
import glob
import logging

from argparse import Namespace

from maccarone.openai import CachedChatAPI
from maccarone.preprocessor import preprocess_maccarone

logger = logging.getLogger(__name__)

def preprocess(
        mn_path: str,
        print_: bool,
        write: bool,
        rewrite: bool,
        suffix: str,
    ) -> None:
    # produce Python source
    logger.info("preprocessing %s", mn_path)

    cache_path = mn_path.replace(suffix, ".mn.json")
    chat_api = CachedChatAPI(cache_path)

    #<<mn_source = read mn_path>>

    py_source = preprocess_maccarone(mn_source, chat_api)

    if write:
        #<<py_path = regex replace mn_path: f"{suffix}$" -> ".py">>

        logger.info("writing %s", py_path)

        if py_path == mn_path:
            raise ValueError("won't overwrite input file", mn_path)
    elif rewrite:
        py_path = mn_path
    else:
        py_path = None

    #<<write py_source to py_path if not None>>

    if print_:
        print(py_source, end="")

def main(path: str, print_: bool, write: bool, rewrite: bool, suffix: str) -> None:
    """Preprocess files with Maccarone snippets."""

    if os.path.isdir(path):
        mn_files = glob.glob(
            os.path.join(path, f"**/*{suffix}"),
            recursive=True,
        )
    else:
        mn_files = [path]

    #<<preprocess mn_files>>

def parse_args() -> Namespace:
    #<<
    # get args for main() and return; use argparse
    # set the `print_` var for `--print`
    # default suffix: ".mn.py"
    #>>

def script_main():
    logging.basicConfig(level=logging.INFO)

    return main(**vars(parse_args()))

if __name__ == "__main__":
    script_main()




from logging import getLogger

logger = getLogger(__name__)

if not __package__.startswith("stale_"):
    try:
        import stale_maccarone
    except ImportError:
        logger.debug("no stale_maccarone; assuming source is already preprocessed")
    else:
        stale_maccarone.enable(
            py_string_matching=False,
            include_pattern="maccarone.*",
        )

from maccarone.loader import enable
